from functools import wraps

from amino import env


def unite(f):
    @wraps(f)
    def wrapper(self):
        def go(unite):
            self.vim.options.amend_l('rtp', [unite])
            self.vim.cmd('source {}/plugin/*.vim'.format(unite))
            self.vim.cmd('source {}/plugin/unite/*.vim'.format(unite))
            self.vim.cmd('source {}/syntax/*.vim'.format(unite))
            return f(self, unite)
        env['UNITE_DIR'] % go
    return wrapper

__all__ = ('unite',)



import abc
import asyncio
from typing import Union

import neovim

from ribosome.nvim import NvimFacade
from ribosome.machine import StateMachine
from ribosome.logging import nvim_logging


class NvimPlugin(object):

    def __init__(self, nvim: Union[NvimFacade, neovim.Nvim]) -> None:
        if isinstance(nvim, neovim.Nvim):
            nvim = NvimFacade(nvim, self.name)
        self.vim = nvim
        self.setup_logging()
        self.setup_asyncio()

    def setup_logging(self):
        nvim_logging(self.vim)

    def setup_asyncio(self):
        ''' this ensures that the child process watcher is running on
        the main thread, which is necessary for subprocess to be waited
        for, as python can only catch SIG_CHLD on the main thread.
        additionally, an event loop must be executed on the main thread
        during waiting.
        '''
        asyncio.get_child_watcher()

    @property
    def loop(self):
        return self.vim.loop

    @property
    def name(self):
        return 'ribosome'


class NvimStatePlugin(NvimPlugin):

    @abc.abstractproperty
    def state(self) -> StateMachine:
        ...


__all__ = ('NvimPlugin', 'NvimStatePlugin')



from ribosome.test.integration import (IntegrationSpec, VimIntegrationSpec,
                                     PluginIntegrationSpec)

__all__ = ('VimIntegrationSpec', 'IntegrationSpec', 'PluginIntegrationSpec')



import neovim

from amino import List, Map, L, __, _


def mk_unite_candidates(Unite):
    def decorator(name):
        handler = getattr(Unite, '{}_candidates'.format(name))
        def uc_wrap(f):
            @neovim.function(handler, sync=True)
            def f_wrap(*a, **kw):
                return f(*a, **kw) / (lambda a: {'word': a})
            return f_wrap
        return uc_wrap
    return decorator


def _unite_word(args):
    first = List.wrap(args).lift(0)
    candidates = (
        first.to_list.join
        if first.exists(lambda a: isinstance(a, list))
        else first
    )
    return candidates / Map // __.get('word')


def mk_unite_action(Unite):
    def decorator(name):
        handler = getattr(Unite, name)
        def uc_wrap(f):
            @neovim.function(handler)
            def f_wrap(self, args):
                _unite_word(args) / L(f)(self, _) % self.state.send
            return f_wrap
        return uc_wrap
    return decorator

__all__ = ('mk_unite_candidates', 'mk_unite_action')



from typing import Callable, Any
from contextlib import contextmanager

from flexmock import flexmock

from amino import may, Maybe, List
from amino.test import Spec

import ribosome
from ribosome.nvim import Buffer, Tab, Window
from ribosome import NvimFacade
from ribosome.nvim.components import Options, Vars


class MockOptions(Options):

    def set(self, *a, **kw):
        pass


class MockVars(Vars):

    def __init__(self, vim) -> None:
        super().__init__(vim)
        self.vars = dict()
        self.prefix = vim.prefix

    @may
    def __call__(self, name: str) -> Maybe[str]:  # type: ignore
        v = self.vars.get(name)
        if v is None:
            ignore_names = ['_machine', '_message',
                            '{}__message'.format(self.prefix),
                            '{}__machine'.format(self.prefix),
                            ]
            if name not in ignore_names:
                self.log.error('variable not found: {}'.format(name))
        return v

    def set(self, name, value):
        self.vars[name] = value


class MockNvim(object):

    def __init__(self, prefix) -> None:
        self.prefix = prefix
        self._vars = MockVars(self)
        self._options = MockOptions(self)

    @property
    def window(self):
        return MockWindow(self.vim, None, self.prefix)

    @property
    def buffer(self):
        return MockBuffer(self.vim, None, self.prefix)

    @property
    def tab(self):
        return MockTab(self.vim, None, self.prefix)

    def cmd(self, *a, **kw):
        pass


class MockTab(MockNvim, Tab):
    pass


class MockWindow(MockNvim, Window):
    pass


class MockBuffer(MockNvim, Buffer):
    pass


class MockNvimFacade(MockNvim, NvimFacade):

    def __init__(self, prefix):
        MockNvim.__init__(self, prefix)
        self.target = self

    @property
    def windows(self):
        return List()

    def switch_root(self, root):
        pass

    def async(self, f: Callable[['NvimFacade'], Any]):
        return f(self)

    @contextmanager
    def main_event_loop(self):
        yield None

    def reload_windows(self):
        pass


class MockNvimSpec(Spec):

    def __init__(self, prefix):
        super().__init__()
        self.prefix = prefix

    def setup(self):
        super().setup()
        ribosome.in_vim = False
        self.vim = MockNvimFacade(self.prefix)
        self.vim_mock = flexmock(self.vim)

__all__ = ('MockNvimSpec', 'MockNvimFacade')



from amino import Right, List, __, Map
from amino.test import later

from ribosome.test import PluginIntegrationSpec

from integration._support.facade import FacadeTestPlugin


class VimSpec(PluginIntegrationSpec):

    @property
    def _prefix(self):
        return 'ribosome'

    @property
    def plugin_class(self):
        return Right(FacadeTestPlugin)

    def _last_output(self, content):
        later(lambda: self._log_out.last.should.contain(content), timeout=1)

    def vars(self):
        vname = List.random_string()
        content = [List.random_string()]
        self.vim.cmd_sync('Go')
        self.vim.buffer.vars.set_p(vname, content)
        later(lambda: self.vim.buffer.vars.pl(vname).should.contain(content))
        var = (self.vim.call('AllVars') / Map //
               __.get('b:{}_{}'.format(self._prefix, vname)))
        var.should.contain(str(content))
        self.vim.vars.p(vname).should.be.empty

__all__ = ('VimSpec',)
